div.menuBar,
span.menuBar,
div.menuBar a.menuButton,
div.menu,
div.menu a.menuItem {
  font-size: 9pt;
  font-style: normal;
  font-weight: normal;
  color: #000000;
}

div.menuBar, span.menuBar {
  background-color: #f7f7ff; /* #eeeeff;  #d0d0d0; */
  padding: 2pt 2pt 2pt 2pt;
  margin: 2pt 2pt 2pt 2pt;
  text-align: left;
}

div.menuBar a.menuButton, span.menuBar, span.menuButton {
  background-color: #eeeeff;
  border: #aaaaff solid 1px;
  color: darkblue;
  cursor: default;
  left: 0px;
  margin: 1px;
  padding: 2px 6px 2px 6px;
  position: relative;
  text-decoration: none;
  font-weight: normal;
  top: 0px;
  z-index: 100;
}

div.menuBar a.menuButton:hover {
  background-color: #bbbbff;
  color: #000000;
}

div.menuBar a.menuButtonActive,
div.menuBar a.menuButtonActive:hover {
  background-color: #bbbbff;
  border-color: #aaaaff;
  left: 1px;
  top: 1px;
}

div.menu {
  background-color: #eeeeff; /* #d0d0d0; */
  border: 1px solid #aaaaff;
  left: 0px;
  padding: 0px 1px 1px 0px;
  position: absolute;
  top: 0px;
  visibility: hidden;
  z-index: 101;
  max-height:250pt;
  overflow-y:auto;
  overflow-x:hidden;
  /*  overflow: auto; */
}

div.menu a.menuItem {
  color: #000000;
  cursor: default;
  display: block;
  padding: 3px 1em;
  text-decoration: none;
  white-space: nowrap;
}

div.menu a.menuItem:hover, div.menu a.menuItemHighlight {
  background-color: #bbbbff;
  color: black;
}

div.menu a.menuItem span.menuItemText {}

span.menuItemArrow {
  margin-right: -.3em;
  margin-left: 0.75em;
  font-size: 7pt;
  color: #aaaaff;
}

div.menu div.menuItemSep {
  border-top: 1px solid #909090;
  border-bottom: 1px solid #f0f0f0;
  margin: 4px 2px;
}
