package PMLTQ::ParserError;
our $AUTHORITY = 'cpan:MICHALS';
{
  $PMLTQ::ParserError::VERSION = '0.8.3';
}

# ABSTRACT: Wrapper for parsing error

1; # End of PMLTQ::ParserError

__END__

=pod

=encoding UTF-8

=head1 NAME

PMLTQ::ParserError - Wrapper for parsing error

=head1 VERSION

version 0.8.3

=head1 AUTHORS

=over 4

=item *

Petr Pajas <pajas@ufal.mff.cuni.cz>

=item *

Jan Štěpánek <stepanek@ufal.mff.cuni.cz>

=item *

Michal Sedlák <sedlak@ufal.mff.cuni.cz>

=back

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2014 by Institute of Formal and Applied Linguistics (http://ufal.mff.cuni.cz).

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
